// encr (c) 2009 Matthias Withopf

extern "C"
  {
    #include "tomcrypt.h"
  };
#include <sys/stat.h>

const unsigned int KeySize = 64;

bool Pwd2Key(char *Pwd,unsigned int SaltLo,unsigned int SaltHi,unsigned char *Key)
{
  bool Result = false;
  if (Pwd && Pwd[0] && Key)
    {
      unsigned int PwdLen = strlen(Pwd);
      const unsigned int MaxInDataSize = 1024;
      unsigned char      InData[MaxInDataSize];
      unsigned int       InDataSize;
      const unsigned int OutDataSize = KeySize;
      unsigned char      OutData[OutDataSize];
      Result = true;
      for (int i = 0;i < 10000;++i)
        {
          if (!i)
            {
              unsigned int l = PwdLen;
              if (l > MaxInDataSize)
                l = MaxInDataSize;
              memcpy(InData,Pwd,l);
              InDataSize = l;
            }
          else
            {
              memcpy(InData,OutData,OutDataSize);
              InDataSize = OutDataSize;
              unsigned int Max = MaxInDataSize - InDataSize;
              unsigned int l = PwdLen;
              if (l > Max)
                l = Max;
              memcpy(InData + InDataSize,Pwd,l);
              InDataSize += l;
            }
          unsigned int Rem     = MaxInDataSize - InDataSize;
          unsigned int ReqSize = sizeof SaltLo + sizeof SaltHi;
          if (Rem >= ReqSize)
            {
              unsigned char *p = (unsigned char *)(InData + InDataSize);
              *p++ = (unsigned char)(SaltLo >> 0);
              *p++ = (unsigned char)(SaltLo >> 8);
              *p++ = (unsigned char)(SaltLo >> 16);
              *p++ = (unsigned char)(SaltLo >> 24);
              *p++ = (unsigned char)(SaltHi >> 0);
              *p++ = (unsigned char)(SaltHi >> 8);
              *p++ = (unsigned char)(SaltHi >> 16);
              *p++ = (unsigned char)(SaltHi >> 24);
              InDataSize += ReqSize;
            }
          unsigned char hash_buffer[MAXBLOCKSIZE];
          hash_state    md;
          whirlpool_desc.init(&md);
          whirlpool_desc.process(&md,InData,InDataSize);
          whirlpool_desc.done(&md,hash_buffer);
          memcpy(OutData,hash_buffer,OutDataSize);
        }
      if (Result)
        memcpy(Key,OutData,KeySize);
    }
  return Result;
}

bool EncryptFile(char *InFileName,char *OutFileName,char *Pwd,unsigned int SaltLo,unsigned int SaltHi)
{
  bool Result = false;
  unsigned char Key[KeySize];
  if (Pwd2Key(Pwd,SaltLo,SaltHi,Key))
    {
      FILE *F = fopen(InFileName,"rb");
      if (!F)
        printf("Error opening input file '%s', error code %u\n",InFileName,errno);
      else
        {
          struct _stat SB;
          if (_stat(InFileName,&SB))
            printf("Error getting file size for file '%s', error code %u\n",InFileName,errno);
          else
            {
              size_t InFileSize = SB.st_size;
              FILE *F2 = fopen(OutFileName,"wb");
              if (!F2)
                printf("Error creating output file '%s', error code %u\n",OutFileName,errno);
              else
                {
                  unsigned char *Key1 = Key + 0;
                  unsigned char *Key2 = Key + KeySize / 2;

                  register_cipher(&aes_desc);
                  int idx;
                  if ((idx = find_cipher("aes")) != -1)
                    {
                      const unsigned int AESBlockSize = 16;
                      unsigned char Block1[AESBlockSize],Block2[AESBlockSize];
                      for (int i = 0;i < AESBlockSize;++i)
                        Block1[i] = (unsigned char)(i ^ 0xA5);

                      symmetric_key skey;
                      int err = cipher_descriptor[idx].setup(Key1,KeySize / 2,0,&skey);
                      if (err == CRYPT_OK)
                        {
                          err = cipher_descriptor[idx].ecb_encrypt(Block1,Block2,&skey);
                          if (err == CRYPT_OK)
                            {
                              unsigned char Header[24];
                              memset(Header,0,sizeof Header);
                              Header[ 0] = 'E';
                              Header[ 1] = 'F';
                              Header[ 2] = Block2[0];
                              Header[ 3] = Block2[1];
                              Header[ 8] = (unsigned char)(SaltLo >> 0);
                              Header[ 9] = (unsigned char)(SaltLo >> 8);
                              Header[10] = (unsigned char)(SaltLo >> 16);
                              Header[11] = (unsigned char)(SaltLo >> 24);
                              Header[12] = (unsigned char)(SaltHi >> 0);
                              Header[13] = (unsigned char)(SaltHi >> 8);
                              Header[14] = (unsigned char)(SaltHi >> 16);
                              Header[15] = (unsigned char)(SaltHi >> 24);
                              Header[16] = (unsigned char)(InFileSize >> 0);
                              Header[17] = (unsigned char)(InFileSize >> 8);
                              Header[18] = (unsigned char)(InFileSize >> 16);
                              Header[19] = (unsigned char)(InFileSize >> 24);

                              size_t WriteSize = fwrite(Header,1,sizeof Header,F2);
                              if (WriteSize != sizeof Header)
                                printf("Error writing file '%s' (%u), error code %u\n",OutFileName,WriteSize,errno);
                              else
                                {
                                  const unsigned int EncryptionBlockSize = 1024;
                                  unsigned char *Buf  = new unsigned char[EncryptionBlockSize];
                                  unsigned char *Buf2 = new unsigned char[EncryptionBlockSize];
                                  if (!Buf || !Buf2)
                                    printf("Out of memory!\n");
                                  else
                                    {
                                      Result = true;
                                      unsigned char T[16];
                                      memset(T,0,sizeof T);
                                      ulong64 CurBlock = 0;
                                      for (size_t Size = InFileSize;Result && Size;++CurBlock)
                                        {
                                          size_t CurSize = Size;
                                          if (CurSize > EncryptionBlockSize)
                                            CurSize = EncryptionBlockSize;
                                          size_t CurSize2 = CurSize;
                                          if (CurSize2 < 16)
                                            CurSize2 = 16;
                                          memset(Buf,0,EncryptionBlockSize);
                                          size_t ReadSize = fread(Buf,1,CurSize2,F);
                                          if (ReadSize != CurSize2)
                                            {
                                              printf("Error reading file '%s' (%u), error code %u\n",InFileName,ReadSize,errno);
                                              Result = false;
                                              break;
                                            }
                                          symmetric_xts xts;
                                          err = xts_start(idx,Key1,Key2,KeySize / 2,0,&xts);
                                          if (err != CRYPT_OK)
                                            {
                                              Result = false;
                                              break;
                                            }
                                          STORE64L(CurBlock,T);
                                          err = xts_encrypt(Buf,CurSize2,Buf2,T,&xts);
                                          if (err != CRYPT_OK)
                                            {
                                              Result = false;
                                              break;
                                            }
                                          WriteSize = fwrite(Buf2,1,CurSize2,F2);
                                          if (WriteSize != CurSize2)
                                            printf("Error writing file '%s' (%u), error code %u\n",OutFileName,WriteSize,errno);

                                          xts_done(&xts);
                                          Size -= CurSize;
                                        }
                                    }
                                  delete [] Buf;
                                  delete [] Buf2;
                                }
                            }
                          cipher_descriptor[idx].done(&skey);
                        }
                    }
                  fclose(F2);
                }
            }
          fclose(F);
        }
    }
  return Result;
}

int main(int argc,char *argv[])
{
  int ExitCode;
  char        *InFileName = NULL;
  char        *Pwd        = NULL;
  unsigned int SaltLo,SaltHi;
  if (argc != 5)
    {
      printf("Usage: %s <InFile> <Password> <SaltLo> <SaltHi>\n",argv[0]);
      ExitCode = 10;
    }
  else
    {
      InFileName = argv[1];
      Pwd        = argv[2];
      SaltLo     = atoi(argv[3]);
      SaltHi     = atoi(argv[4]);
      char OutFileName[256];
      strncpy(OutFileName,InFileName,sizeof OutFileName);
      OutFileName[sizeof OutFileName - 1] = '\0';
      strncat(OutFileName,"-encrypted",sizeof OutFileName);
      OutFileName[sizeof OutFileName - 1] = '\0';

      if (!EncryptFile(InFileName,OutFileName,Pwd,SaltLo,SaltHi))
        {
          printf("Error encrypting file '%s'!\n",InFileName);
          ExitCode = 15;
        }
      else
        ExitCode = 0;
    }
  return ExitCode;
}

